import { encode } from 'vlq';

function Chunk ( start, end, content ) {
	this.start = start;
	this.end = end;
	this.original = content;

	this.intro = '';
	this.outro = '';

	this.content = content;
	this.storeName = false;
	this.edited = false;
}

Chunk.prototype = {
	append: function append ( content ) {
		this.outro += content;
	},

	clone: function clone () {
		var chunk = new Chunk( this.start, this.end, this.original );
		chunk.content = this.content;
		chunk.storeName = this.storeName;
		chunk.edited = this.edited;

		return chunk;
	},

	edit: function edit ( content, storeName ) {
		this.content = content;
		this.storeName = storeName;

		this.edited = true;

		return this;
	},

	prepend: function prepend ( content ) {
		this.intro = content + this.intro;
	},

	split: function split ( index ) {
		if ( index === this.start ) return this;

		var sliceIndex = index - this.start;

		var originalBefore = this.original.slice( 0, sliceIndex );
		var originalAfter = this.original.slice( sliceIndex );

		this.original = originalBefore;

		var newChunk = new Chunk( index, this.end, originalAfter );
		newChunk.outro = this.outro;
		this.outro = '';

		this.end = index;

		if ( this.edited ) {
			if ( this.content.length ) throw new Error( ("Cannot split a chunk that has already been edited (\"" + (this.original) + "\")") );

			// zero-length edited chunks are a special case (overlapping replacements)
			newChunk.edit( '', false );
			this.content = '';
		} else {
			this.content = originalBefore;
		}

		return newChunk;
	},

	toString: function toString () {
		return this.intro + this.content + this.outro;
	}
};

var _btoa;

if ( typeof window !== 'undefined' && typeof window.btoa === 'function' ) {
	_btoa = window.btoa;
} else if ( typeof Buffer === 'function' ) {
	/* global Buffer */
	_btoa = function ( str ) { return new Buffer( str ).toString( 'base64' ); };
} else {
	throw new Error( 'Unsupported environment: `window.btoa` or `Buffer` should be supported.' );
}

var btoa = _btoa;

function SourceMap ( properties ) {
	this.version = 3;

	this.file           = properties.file;
	this.sources        = properties.sources;
	this.sourcesContent = properties.sourcesContent;
	this.names          = properties.names;
	this.mappings       = properties.mappings;
}

SourceMap.prototype = {
	toString: function toString () {
		return JSON.stringify( this );
	},

	toUrl: function toUrl () {
		return 'data:application/json;charset=utf-8;base64,' + btoa( this.toString() );
	}
};

function guessIndent ( code ) {
	var lines = code.split( '\n' );

	var tabbed = lines.filter( function ( line ) { return /^\t+/.test( line ); } );
	var spaced = lines.filter( function ( line ) { return /^ {2,}/.test( line ); } );

	if ( tabbed.length === 0 && spaced.length === 0 ) {
		return null;
	}

	// More lines tabbed than spaced? Assume tabs, and
	// default to tabs in the case of a tie (or nothing
	// to go on)
	if ( tabbed.length >= spaced.length ) {
		return '\t';
	}

	// Otherwise, we need to guess the multiple
	var min = spaced.reduce( function ( previous, current ) {
		var numSpaces = /^ +/.exec( current )[0].length;
		return Math.min( numSpaces, previous );
	}, Infinity );

	return new Array( min + 1 ).join( ' ' );
}

function getLocator ( source ) {
	var originalLines = source.split( '\n' );

	return function locate ( index ) {
		var len = originalLines.length;

		var lineStart = 0;

		for ( var i = 0; i < len; i += 1 ) {
			var line = originalLines[i];
			var lineEnd =  lineStart + line.length + 1; // +1 for newline

			if ( lineEnd > index ) return { line: i, column: index - lineStart };

			lineStart = lineEnd;
		}
	};
}

function encodeMappings ( original, intro, chunks, hires, sourcemapLocations, sourceIndex, offsets, names ) {
	var rawLines = [];

	var generatedCodeLine = intro.split( '\n' ).length - 1;
	var rawSegments = rawLines[ generatedCodeLine ] = [];

	var generatedCodeColumn = 0;

	var locate = getLocator( original );

	function addUneditedChunk ( chunk ) {
		var originalCharIndex = chunk.start;
		var first = true;

		var ref = locate( originalCharIndex ), line = ref.line, column = ref.column;

		while ( originalCharIndex < chunk.end ) {
			if ( hires || first || sourcemapLocations[ originalCharIndex ] ) {
				rawSegments.push({
					generatedCodeLine: generatedCodeLine,
					generatedCodeColumn: generatedCodeColumn,
					sourceCodeLine: line,
					sourceCodeColumn: column,
					sourceCodeName: -1,
					sourceIndex: sourceIndex
				});
			}

			if ( original[ originalCharIndex ] === '\n' ) {
				line += 1;
				column = 0;
				generatedCodeLine += 1;
				rawLines[ generatedCodeLine ] = rawSegments = [];
				generatedCodeColumn = 0;
			} else {
				column += 1;
				generatedCodeColumn += 1;
			}

			originalCharIndex += 1;
			first = false;
		}
	}

	for ( var i = 0; i < chunks.length; i += 1 ) {
		var chunk = chunks[i];
		var ref = locate( chunk.start ), line = ref.line, column = ref.column;

		if ( chunk.edited ) {
			if ( i > 0 || chunk.content.length ) {
				rawSegments.push({
					generatedCodeLine: generatedCodeLine,
					generatedCodeColumn: generatedCodeColumn,
					sourceCodeLine: line,
					sourceCodeColumn: column,
					sourceCodeName: chunk.storeName ? names.indexOf( chunk.original ) : -1,
					sourceIndex: sourceIndex
				});
			}

			var lines = chunk.content.split( '\n' );
			var lastLine = lines.pop();

			if ( lines.length ) {
				generatedCodeLine += lines.length;
				rawLines[ generatedCodeLine ] = rawSegments = [];
				generatedCodeColumn = lastLine.length;
			} else {
				generatedCodeColumn += lastLine.length;
			}

			lines = chunk.original.split( '\n' );
			lastLine = lines.pop();

			if ( lines.length ) {
				line += lines.length;
				column = lastLine.length;
			} else {
				column += lastLine.length;
			}
		} else {
			addUneditedChunk( chunk );
		}
	}

	offsets.sourceIndex = offsets.sourceIndex || 0;
	offsets.sourceCodeLine = offsets.sourceCodeLine || 0;
	offsets.sourceCodeColumn = offsets.sourceCodeColumn || 0;
	offsets.sourceCodeName = offsets.sourceCodeName || 0;

	var encoded = rawLines.map( function ( segments ) {
		var generatedCodeColumn = 0;

		return segments.map( function ( segment ) {
			var arr = [
				segment.generatedCodeColumn - generatedCodeColumn,
				segment.sourceIndex - offsets.sourceIndex,
				segment.sourceCodeLine - offsets.sourceCodeLine,
				segment.sourceCodeColumn - offsets.sourceCodeColumn
			];

			generatedCodeColumn = segment.generatedCodeColumn;
			offsets.sourceIndex = segment.sourceIndex;
			offsets.sourceCodeLine = segment.sourceCodeLine;
			offsets.sourceCodeColumn = segment.sourceCodeColumn;

			if ( ~segment.sourceCodeName ) {
				arr.push( segment.sourceCodeName - offsets.sourceCodeName );
				offsets.sourceCodeName = segment.sourceCodeName;
			}

			return encode( arr );
		}).join( ',' );
	}).join( ';' );

	return encoded;
}

function getRelativePath ( from, to ) {
	var fromParts = from.split( /[\/\\]/ );
	var toParts = to.split( /[\/\\]/ );

	fromParts.pop(); // get dirname

	while ( fromParts[0] === toParts[0] ) {
		fromParts.shift();
		toParts.shift();
	}

	if ( fromParts.length ) {
		var i = fromParts.length;
		while ( i-- ) fromParts[i] = '..';
	}

	return fromParts.concat( toParts ).join( '/' );
}

var toString = Object.prototype.toString;

function isObject ( thing ) {
	return toString.call( thing ) === '[object Object]';
}

function findIndex ( array, fn ) {
	for ( var i = 0; i < array.length; i += 1 ) {
		if ( fn( array[i], i ) ) return i;
	}

	return -1;
}

function find ( array, fn ) {
	return array[ findIndex( array, fn ) ];
}

var warned = false;

function MagicString ( string, options ) {
	if ( options === void 0 ) options = {};

	var chunk = new Chunk( 0, string.length, string );

	Object.defineProperties( this, {
		original:              { writable: true, value: string },
		outro:                 { writable: true, value: '' },
		intro:                 { writable: true, value: '' },
		chunks:                { writable: true, value: [ chunk ] },
		moves:                 { writable: true, value: [] },
		filename:              { writable: true, value: options.filename },
		indentExclusionRanges: { writable: true, value: options.indentExclusionRanges },
		sourcemapLocations:    { writable: true, value: {} },
		storedNames:           { writable: true, value: {} },
		indentStr:             { writable: true, value: guessIndent( string ) }
	});
}

MagicString.prototype = {
	addSourcemapLocation: function addSourcemapLocation ( char ) {
		this.sourcemapLocations[ char ] = true;
	},

	append: function append ( content ) {
		if ( typeof content !== 'string' ) throw new TypeError( 'outro content must be a string' );

		this.outro += content;
		return this;
	},

	clone: function clone () {
		var cloned = new MagicString( this.original, { filename: this.filename });

		cloned.chunks = this.chunks.map( function ( chunk ) { return chunk.clone(); } );

		if ( this.indentExclusionRanges ) {
			cloned.indentExclusionRanges = typeof this.indentExclusionRanges[0] === 'number' ?
				[ this.indentExclusionRanges[0], this.indentExclusionRanges[1] ] :
				this.indentExclusionRanges.map( function ( range ) { return [ range.start, range.end ]; } );
		}

		Object.keys( this.sourcemapLocations ).forEach( function ( loc ) {
			cloned.sourcemapLocations[ loc ] = true;
		});

		return cloned;
	},

	generateMap: function generateMap ( options ) {
		options = options || {};

		var names = Object.keys( this.storedNames );

		return new SourceMap({
			file: ( options.file ? options.file.split( /[\/\\]/ ).pop() : null ),
			sources: [ options.source ? getRelativePath( options.file || '', options.source ) : null ],
			sourcesContent: options.includeContent ? [ this.original ] : [ null ],
			names: names,
			mappings: this.getMappings( options.hires, 0, {}, names )
		});
	},

	getIndentString: function getIndentString () {
		return this.indentStr === null ? '\t' : this.indentStr;
	},

	getMappings: function getMappings ( hires, sourceIndex, offsets, names ) {
		return encodeMappings( this.original, this.intro, this.chunks, hires, this.sourcemapLocations, sourceIndex, offsets, names );
	},

	indent: function indent ( indentStr, options ) {
		var this$1 = this;

		var pattern = /^[^\r\n]/gm;

		if ( isObject( indentStr ) ) {
			options = indentStr;
			indentStr = undefined;
		}

		indentStr = indentStr !== undefined ? indentStr : ( this.indentStr || '\t' );

		if ( indentStr === '' ) return this; // noop

		options = options || {};

		// Process exclusion ranges
		var isExcluded = {};

		if ( options.exclude ) {
			var exclusions = typeof options.exclude[0] === 'number' ? [ options.exclude ] : options.exclude;
			exclusions.forEach( function ( exclusion ) {
				for ( var i = exclusion[0]; i < exclusion[1]; i += 1 ) {
					isExcluded[i] = true;
				}
			});
		}

		var shouldIndentNextCharacter = options.indentStart !== false;
		var replacer = function ( match ) {
			if ( shouldIndentNextCharacter ) return ("" + indentStr + match);
			shouldIndentNextCharacter = true;
			return match;
		};

		this.intro = this.intro.replace( pattern, replacer );

		var chunkIndex;
		var charIndex = 0;

		for ( chunkIndex = 0; chunkIndex < this$1.chunks.length; chunkIndex += 1 ) { // can't cache this.chunks.length, it may change
			var chunk = this$1.chunks[ chunkIndex ];
			var end = chunk.end;

			if ( chunk.edited ) {
				if ( !isExcluded[ charIndex ] ) {
					chunk.content = chunk.content.replace( pattern, replacer );

					if ( chunk.content.length ) {
						shouldIndentNextCharacter = chunk.content[ chunk.content.length - 1 ] === '\n';
					}
				}
			} else {
				charIndex = chunk.start;

				while ( charIndex < end ) {
					if ( !isExcluded[ charIndex ] ) {
						var char = this$1.original[ charIndex ];

						if ( char === '\n' ) {
							shouldIndentNextCharacter = true;
						} else if ( char !== '\r' && shouldIndentNextCharacter ) {
							shouldIndentNextCharacter = false;

							var indentation = new Chunk( charIndex, charIndex, '' ).edit( indentStr, false );
							var remainder = chunk.split( charIndex );

							if ( charIndex === chunk.start ) {
								this$1.chunks.splice( chunkIndex, 0, indentation );
								chunkIndex += 1;
							} else {
								this$1.chunks.splice( chunkIndex + 1, 0, indentation, remainder );
								chunkIndex += 2;
							}

							chunk = remainder;
						}
					}

					charIndex += 1;
				}
			}

			charIndex = chunk.end;
		}

		this.outro = this.outro.replace( pattern, replacer );

		return this;
	},

	insert: function insert ( index, content ) {
		if ( typeof content !== 'string' ) throw new TypeError( 'inserted content must be a string' );

		this._split( index );

		var next = findIndex( this.chunks, function ( chunk ) { return chunk.original.length && chunk.start === index; } );
		if ( !~next ) next = this.chunks.length;

		var newChunk = new Chunk( index, index, '' ).edit( content, false );

		this.chunks.splice( next, 0, newChunk );
		return this;
	},

	insertAfter: function insertAfter ( index, content ) {
		if ( typeof content !== 'string' ) throw new TypeError( 'inserted content must be a string' );

		this._split( index );

		var chunk = find( this.chunks, function ( chunk ) { return chunk.end === index; } );

		if ( chunk ) {
			chunk.append( content );
		} else {
			this.intro += content;
		}

		return this;
	},

	insertBefore: function insertBefore ( index, content ) {
		if ( typeof content !== 'string' ) throw new TypeError( 'inserted content must be a string' );

		this._split( index );

		var chunk = find( this.chunks, function ( chunk ) { return chunk.start === index; } );

		if ( chunk ) {
			chunk.prepend( content );
		} else {
			this.outro += content;
		}

		return this;
	},

	// get current location of character in original string
	locate: function locate () {
		throw new Error( 'magicString.locate is deprecated' );
	},

	locateOrigin: function locateOrigin () {
		throw new Error( 'magicString.locateOrigin is deprecated' );
	},

	move: function move ( start, end, index ) {
		if ( index >= start && index <= end ) throw new Error( 'Cannot move a selection inside itself' );

		this._split( start );
		this._split( end );
		this._split( index );

		var firstIndex = findIndex( this.chunks, function ( chunk ) { return chunk.start === start; } );
		var lastIndex = findIndex( this.chunks, function ( chunk ) { return chunk.end === end; } );

		var toMove = this.chunks.splice( firstIndex, lastIndex + 1 - firstIndex );

		var insertionIndex = findIndex( this.chunks, function ( chunk ) { return chunk.original.length && chunk.start === index; } );
		if ( !~insertionIndex ) insertionIndex = this.chunks.length;

		this.chunks.splice.apply( this.chunks, [ insertionIndex, 0 ].concat( toMove ) );

		return this;
	},

	overwrite: function overwrite ( start, end, content, storeName ) {
		if ( typeof content !== 'string' ) {
			throw new TypeError( 'replacement content must be a string' );
		}

		this._split( start );
		this._split( end );

		if ( storeName ) {
			var original = this.original.slice( start, end );
			this.storedNames[ original ] = true;
		}

		var firstIndex = findIndex( this.chunks, function ( chunk ) { return chunk.start === start && chunk.original.length; } );
		if ( !~firstIndex ) firstIndex = this.chunks.length;
		var lastIndex = findIndex( this.chunks, function ( chunk ) { return chunk.end === end; } );

		var firstChunk = this.chunks[ firstIndex ];

		var newChunk = new Chunk( start, end, this.original.slice( start, end ) ).edit( content, storeName );
		if ( firstChunk ) newChunk.intro = firstChunk.intro;

		this.chunks.splice( firstIndex, lastIndex + 1 - firstIndex, newChunk );
		return this;
	},

	prepend: function prepend ( content ) {
		if ( typeof content !== 'string' ) throw new TypeError( 'outro content must be a string' );

		this.intro = content + this.intro;
		return this;
	},

	remove: function remove ( start, end ) {
		var this$1 = this;

		while ( start < 0 ) start += this$1.original.length;
		while ( end < 0 ) end += this$1.original.length;

		if ( start === end ) return this;

		if ( start < 0 || end > this.original.length ) throw new Error( 'Character is out of bounds' );
		if ( start > end ) throw new Error( 'end must be greater than start' );

		var firstIndex = findIndex( this.chunks, function ( chunk ) { return chunk.start <= start && chunk.end > start; } );
		var chunk = this.chunks[ firstIndex ];

		// if the chunk contains `start`, split
		if ( chunk.start < start ) {
			if ( chunk.edited && chunk.content.length ) throw new Error( ("Cannot remove edited content (\"" + (this.original.slice(start, end)) + "\")") );
			this._split( start );
			firstIndex += 1;
		}

		var lastIndex = findIndex( this.chunks, function ( chunk ) { return chunk.start < end && chunk.end >= end; } );
		chunk = this.chunks[ lastIndex ];

		// if the chunk contains `end`, split
		if ( chunk.start < end ) {
			if ( chunk.edited && chunk.content.length ) throw new Error( ("Cannot remove edited content (\"" + (this.original.slice(start, end)) + "\")") );
			this._split( end );
		}

		lastIndex += 1;

		var newChunk = new Chunk( start, end, this.original.slice( start, end ) ).edit( '', false );
		this.chunks.splice( firstIndex, lastIndex - firstIndex, newChunk );

		return this;
	},

	replace: function replace ( start, end, content ) {
		if ( !warned ) {
			console.warn( 'magicString.replace(...) is deprecated. Use magicString.overwrite(...) instead' ); // eslint-disable-line no-console
			warned = true;
		}

		return this.overwrite( start, end, content );
	},

	slice: function slice ( start, end ) {
		var this$1 = this;
		if ( end === void 0 ) end = this.original.length;

		while ( start < 0 ) start += this$1.original.length;
		while ( end < 0 ) end += this$1.original.length;

		var result = '';

		for ( var i = 0; i < this$1.chunks.length; i += 1 ) {
			var chunk = this$1.chunks[i];

			if ( chunk.end <= start ) continue;
			if ( chunk.start >= end ) break;

			if ( chunk.start < start || chunk.end > end ) {
				if ( chunk.edited ) throw new Error( ("Cannot use replaced characters (" + start + ", " + end + ") as slice anchors") );

				var sliceStart = Math.max( start - chunk.start, 0 );
				var sliceEnd = Math.min( chunk.content.length - ( chunk.end - end ), chunk.content.length );

				result += chunk.content.slice( sliceStart, sliceEnd );
			} else {
				result += chunk.content;
			}
		}

		return result;
	},

	snip: function snip ( start, end ) {
		var clone = this.clone();
		clone.remove( 0, start );
		clone.remove( end, clone.original.length );

		return clone;
	},

	_split: function _split ( index ) {
		// TODO bisect
		var this$1 = this;

		for ( var i = 0; i < this$1.chunks.length; i += 1 ) {
			var chunk = this$1.chunks[i];
			if ( chunk.start === index || chunk.end === index ) return;

			if ( chunk.start < index && chunk.end > index ) {
				var newChunk = chunk.split( index );
				this$1.chunks.splice( i + 1, 0, newChunk );
				return;
			}
		}
	},

	toString: function toString () {
		return this.intro + this.chunks.map( function ( chunk ) { return chunk.toString(); } ).join( '' ) + this.outro;
	},

	trimLines: function trimLines () {
		return this.trim('[\\r\\n]');
	},

	// TODO rewrite these methods, post-refactor
	trim: function trim ( charType ) {
		return this.trimStart( charType ).trimEnd( charType );
	},

	trimEnd: function trimEnd ( charType ) {
		var this$1 = this;

		var rx = new RegExp( ( charType || '\\s' ) + '+$' );

		this.outro = this.outro.replace( rx, '' );
		if ( this.outro.length ) return this;

		do {
			var lastChunk = this$1.chunks[ this$1.chunks.length - 1 ];

			if ( lastChunk.content.length ) {
				var match = rx.exec( lastChunk.content );
				if ( !match ) return this$1;

				if ( lastChunk.edited ) {
					lastChunk.edit( lastChunk.content.slice( 0, match.index ) );
				} else {
					if ( match.index ) {
						lastChunk.split( match.index + lastChunk.start ); // generated chunk is discarded
					} else {
						lastChunk.edit( '' );
					}
				}

				if ( match.index > 0 ) return this$1;
			}

			if ( this$1.chunks.length === 1 ) return this$1;
			this$1.chunks.pop();
		} while ( true );
	},

	trimStart: function trimStart ( charType ) {
		var this$1 = this;

		var rx = new RegExp( '^' + ( charType || '\\s' ) + '+' );

		this.intro = this.intro.replace( rx, '' );
		if ( this.intro.length ) return this;

		do {
			var firstChunk = this$1.chunks[0];

			if ( firstChunk.content.length ) {
				var match = rx.exec( firstChunk.content );
				if ( !match ) return this$1;

				var end = match.index + match[0].length;

				if ( firstChunk.edited ) {
					firstChunk.edit( firstChunk.content.slice( match.index ) );
				} else {
					var newChunk = firstChunk.split( end + firstChunk.start ); // existing chunk is discarded
					this$1.chunks[0] = newChunk;

					continue;
				}

				if ( end < firstChunk.content.length ) return this$1;
			}

			if ( this$1.chunks.length === 1 ) return this$1;
			this$1.chunks.shift();
		} while ( true );
	}
};

var hasOwnProp = Object.prototype.hasOwnProperty;

function Bundle ( options ) {
	if ( options === void 0 ) options = {};

	this.intro = options.intro || '';
	this.separator = options.separator !== undefined ? options.separator : '\n';

	this.sources = [];

	this.uniqueSources = [];
	this.uniqueSourceIndexByFilename = {};
}

Bundle.prototype = {
	addSource: function addSource ( source ) {
		if ( source instanceof MagicString ) {
			return this.addSource({
				content: source,
				filename: source.filename,
				separator: this.separator
			});
		}

		if ( !isObject( source ) || !source.content ) {
			throw new Error( 'bundle.addSource() takes an object with a `content` property, which should be an instance of MagicString, and an optional `filename`' );
		}

		[ 'filename', 'indentExclusionRanges', 'separator' ].forEach( function ( option ) {
			if ( !hasOwnProp.call( source, option ) ) source[ option ] = source.content[ option ];
		});

		if ( source.separator === undefined ) { // TODO there's a bunch of this sort of thing, needs cleaning up
			source.separator = this.separator;
		}

		if ( source.filename ) {
			if ( !hasOwnProp.call( this.uniqueSourceIndexByFilename, source.filename ) ) {
				this.uniqueSourceIndexByFilename[ source.filename ] = this.uniqueSources.length;
				this.uniqueSources.push({ filename: source.filename, content: source.content.original });
			} else {
				var uniqueSource = this.uniqueSources[ this.uniqueSourceIndexByFilename[ source.filename ] ];
				if ( source.content.original !== uniqueSource.content ) {
					throw new Error( ("Illegal source: same filename (" + (source.filename) + "), different contents") );
				}
			}
		}

		this.sources.push( source );
		return this;
	},

	append: function append ( str, options ) {
		this.addSource({
			content: new MagicString( str ),
			separator: ( options && options.separator ) || ''
		});

		return this;
	},

	clone: function clone () {
		var bundle = new Bundle({
			intro: this.intro,
			separator: this.separator
		});

		this.sources.forEach( function ( source ) {
			bundle.addSource({
				filename: source.filename,
				content: source.content.clone(),
				separator: source.separator
			});
		});

		return bundle;
	},

	generateMap: function generateMap ( options ) {
		var this$1 = this;

		var offsets = {};

		var names = [];
		this.sources.forEach( function ( source ) {
			Object.keys( source.content.storedNames ).forEach( function ( name ) {
				if ( !~names.indexOf( name ) ) names.push( name );
			});
		});

		var encoded = (
			getSemis( this.intro ) +
			this.sources.map( function ( source, i ) {
				var prefix = ( i > 0 ) ? ( getSemis( source.separator ) || ',' ) : '';
				var mappings;

				// we don't bother encoding sources without a filename
				if ( !source.filename ) {
					mappings = getSemis( source.content.toString() );
				} else {
					var sourceIndex = this$1.uniqueSourceIndexByFilename[ source.filename ];
					mappings = source.content.getMappings( options.hires, sourceIndex, offsets, names );
				}

				return prefix + mappings;
			}).join( '' )
		);

		return new SourceMap({
			file: ( options.file ? options.file.split( /[\/\\]/ ).pop() : null ),
			sources: this.uniqueSources.map( function ( source ) {
				return options.file ? getRelativePath( options.file, source.filename ) : source.filename;
			}),
			sourcesContent: this.uniqueSources.map( function ( source ) {
				return options.includeContent ? source.content : null;
			}),
			names: names,
			mappings: encoded
		});
	},

	getIndentString: function getIndentString () {
		var indentStringCounts = {};

		this.sources.forEach( function ( source ) {
			var indentStr = source.content.indentStr;

			if ( indentStr === null ) return;

			if ( !indentStringCounts[ indentStr ] ) indentStringCounts[ indentStr ] = 0;
			indentStringCounts[ indentStr ] += 1;
		});

		return ( Object.keys( indentStringCounts ).sort( function ( a, b ) {
			return indentStringCounts[a] - indentStringCounts[b];
		})[0] ) || '\t';
	},

	indent: function indent ( indentStr ) {
		var this$1 = this;

		if ( !arguments.length ) {
			indentStr = this.getIndentString();
		}

		if ( indentStr === '' ) return this; // noop

		var trailingNewline = !this.intro || this.intro.slice( -1 ) === '\n';

		this.sources.forEach( function ( source, i ) {
			var separator = source.separator !== undefined ? source.separator : this$1.separator;
			var indentStart = trailingNewline || ( i > 0 && /\r?\n$/.test( separator ) );

			source.content.indent( indentStr, {
				exclude: source.indentExclusionRanges,
				indentStart: indentStart//: trailingNewline || /\r?\n$/.test( separator )  //true///\r?\n/.test( separator )
			});

			// TODO this is a very slow way to determine this
			trailingNewline = source.content.toString().slice( 0, -1 ) === '\n';
		});

		if ( this.intro ) {
			this.intro = indentStr + this.intro.replace( /^[^\n]/gm, function ( match, index ) {
				return index > 0 ? indentStr + match : match;
			});
		}

		return this;
	},

	prepend: function prepend ( str ) {
		this.intro = str + this.intro;
		return this;
	},

	toString: function toString () {
		var this$1 = this;

		var body = this.sources.map( function ( source, i ) {
			var separator = source.separator !== undefined ? source.separator : this$1.separator;
			var str = ( i > 0 ? separator : '' ) + source.content.toString();

			return str;
		}).join( '' );

		return this.intro + body;
	},

	trimLines: function trimLines () {
		return this.trim('[\\r\\n]');
	},

	trim: function trim ( charType ) {
		return this.trimStart( charType ).trimEnd( charType );
	},

	trimStart: function trimStart ( charType ) {
		var this$1 = this;

		var rx = new RegExp( '^' + ( charType || '\\s' ) + '+' );
		this.intro = this.intro.replace( rx, '' );

		if ( !this.intro ) {
			var source;
			var i = 0;

			do {
				source = this$1.sources[i];

				if ( !source ) {
					break;
				}

				source.content.trimStart( charType );
				i += 1;
			} while ( source.content.toString() === '' ); // TODO faster way to determine non-empty source?
		}

		return this;
	},

	trimEnd: function trimEnd ( charType ) {
		var this$1 = this;

		var rx = new RegExp( ( charType || '\\s' ) + '+$' );

		var source;
		var i = this.sources.length - 1;

		do {
			source = this$1.sources[i];

			if ( !source ) {
				this$1.intro = this$1.intro.replace( rx, '' );
				break;
			}

			source.content.trimEnd( charType );
			i -= 1;
		} while ( source.content.toString() === '' ); // TODO faster way to determine non-empty source?

		return this;
	}
};

function getSemis ( str ) {
	return new Array( str.split( '\n' ).length ).join( ';' );
}

MagicString.Bundle = Bundle;

export default MagicString;
//# sourceMappingURL=magic-string.es6.js.map